#include "FlexSensor.h"
#include "Arduino.h"

FlexSensor::FlexSensor(int pin) {
  this->pin = pin;

  this->active = 1;
  this->calibrated = 0;
  this->continuousValues = 0;
  this->longClickEnabled = 1;
  this->bothDirections = 0;

  this->standardValue = 0;
  this->countSignals = 0;
  this->pSignal = false;
  this->errorState = false;

  this->tolerance = 0;
}

FlexSensor::FlexSensor(int pin, int tolerance) {
  this->pin = pin;

  this->active = 1;
  this->calibrated = 0;
  this->continuousValues = 0;
  this->longClickEnabled = 1;
  this->bothDirections = 0;

  this->standardValue = 0;
  this->countSignals = 0;
  this->pSignal = false;
  this->errorState = false;

  this->tolerance = tolerance;
}


//
//SETTINGS_______________________________________________________________________________________
//
void FlexSensor::setLedPins(int onLedPin, int sleepLedPin) {
  this->onLedPin = onLedPin;
  this->sleepLedPin = sleepLedPin;
}

void FlexSensor::enableLongClick(int state) {
  this->longClickEnabled = state;
  Serial.println("longCl" + String(state));
}

void FlexSensor::enableContinuousValues(int state) {
  this->continuousValues = state;
  Serial.println("contVal" + String(state));
}

void FlexSensor::enableBothDirections(int state) {
  this->bothDirections = state;
  Serial.println("bothDir" + String(state));
}

//
//ACTION_______________________________________________________________________________________
//
void FlexSensor::calibrate() {
  analogWrite(onLedPin, 0);

  long value = 0;
  Serial.println("Calibrating the Sensor. Don't flex it!");
  delay(500);
  for (int i = 0; i < 100; i++) {
    if (i % 24 == 0) {
      analogWrite(sleepLedPin, sleepLedBrightness);
    } else  if (i % 24 == 12) {
      analogWrite(sleepLedPin, 0);
    }
    value += analogRead(this->pin);
    delay(10);
    if (i % 2) {
      Serial.print(".");
    }
  }
  value /= 100;
  this->standardValue = value;
  Serial.println("\nFinished calibrating. Sensor is ready! Average value is: " + String(this->standardValue));
  if (value < 20) {
    Serial.println("$$Check if your sensor is connected!$$");
  }
  analogWrite(onLedPin, onLedBrightness);
  analogWrite(sleepLedPin, 0);
  this->calibrated = 1;
}


void FlexSensor::reset() {
  this->pin = pin;

  this->active = 1;
  this->calibrated = 0;

  this->standardValue = 0;
  this->countSignals = 0;
  this->pSignal = false;
  this->errorState = false;

  Serial.println("reset");

}


int FlexSensor::setActive(int activeState) {

  if (activeState == 0 || activeState == 1) {
    this->active = activeState;
    Serial.println("state" + String(this->active)); //return state value to computer
    if (activeState) {
      analogWrite(onLedPin, onLedBrightness);
      analogWrite(sleepLedPin, 0);
    } else {
      analogWrite(onLedPin, 0);
      analogWrite(sleepLedPin, sleepLedBrightness);
    }
    return 1;
  } else {
    return 0;
  }

}


int FlexSensor::readData() {

  int value = analogRead(this->pin);


  if (this->continuousValues) {
    Serial.println(value - this->standardValue);

  } else {
    analogWrite(sleepLedPin, 0);

    bool currentSignal = false;
    int difference = value - standardValue;
    if (abs(difference) > 200) {
      if (!errorState) {
        Serial.println("$$Mb ur Sensor has bn disc.$$");
        errorState = true;
      }
    }  else {
      errorState = false;
    }

    if (abs(difference) > tolerance) {
      currentSignal = true; //new signal value
    }

    if (currentSignal) { //press started or continued
      countSignals++;
    }

    if (pSignal && !currentSignal) { //pressReleased

      if (countSignals > 2*longThresh) {
        Serial.println(VERYLONG);
        analogWrite(sleepLedPin, sleepLedBrightness);

      } else if (countSignals > longThresh) {
        Serial.println((bothDirections && difference < 0) ? LONGV : LONG);
        analogWrite(sleepLedPin, sleepLedBrightness);

      } else if ( countSignals > shortThresh) {
        Serial.println((bothDirections && difference < 0) ? SHORTV : SHORT);
        analogWrite(sleepLedPin, sleepLedBrightness);
      }

      countSignals = 0; //reset count of signals
    }

    pSignal = currentSignal; //set new previous

  }
}



//
//VALUES_______________________________________________________________________________________
//
int FlexSensor::setTolerance(int tolerance) {
  if (tolerance >= 0 && tolerance <= 2147483647) {
    this->tolerance = tolerance;
    Serial.println("tolerance" + String(this->tolerance)); //return tolerance to computer
    return 1;
  } else {
    return 0;
  }
}

void FlexSensor::setShortThresh(int newThresh) {
  if (newThresh >= 0 && newThresh <= 2147483647) {
    this->shortThresh = newThresh;
    Serial.println("shortTr" + String(this->shortThresh)); //return longlength to computer
    return 1;
  } else {
    return 0;
  }
}

void FlexSensor::setLongThresh(int newThresh) {
  if (newThresh >= 0 && newThresh <= 2147483647) {
    this->longThresh = newThresh;
    Serial.println("longTr" + String(this->longThresh)); //return longlength to computer
    return 1;
  } else {
    return 0;
  }
}




//
//GETTER_______________________________________________________________________________________
//
int FlexSensor::isActive() {
  return this->active;
}

int FlexSensor::isCalibrated() {
  return this->calibrated;
}

int FlexSensor::getTolerance() {
  return tolerance;
}

void FlexSensor::getShortThresh(int) {
  return shortThresh;
}

void FlexSensor::getLongThresh(int) {
  return longThresh;
}
