
package pGUI.core;

import processing.core.*;
import processing.event.*;
import java.awt.EventQueue;
import java.lang.reflect.Method;

import java.lang.reflect.InvocationTargetException;
import pGUI.classes.*;

/*
 * Main component all other components inherit from. 
 * 
 * It provides all basic style attributes and methods. 
 * 
 * 
 */

public abstract class Control {

  // name for the object can be specified manually, useful for distinguishing
  public String Name = "";

  // parent container and frame of object
  protected Control parent;
  protected Frame frame;
  protected PApplet papplet;

  // image that is of the size of the control and contains the looks
  protected PGraphics pg;

  // changedVisuals will be set to true when graphics of this control changed, so
  // parents know if they need to re-render this object.
  // This will happen when update() has been called for this Control.
  // - update() also calls update() for all parent containers
  protected boolean changedVisuals = true;

  /*
   * Status properties
   */
  protected boolean Focusable = true;
  protected boolean Focused = false; // Dont'change focused from outside
  protected boolean stickyFocus = false; // when true than its focused state can't be overridden by other elements
  // requesting focus, only when itself calls blur

  protected boolean Enabled = true;
  protected boolean Visible = true;

  /*
   * Coordinates and size relative to container; won't be changed by
   * flowContainer!
   */
  protected int X;
  protected int Y;
  protected int Z; // z-index of component, only has an effect when component is in a container

  protected int Height;
  protected int Width;

  protected int MinHeight = 1;
  protected int MaxHeight = 1000;
  protected int MinWidth = 1;
  protected int MaxWidth = 1000;

  protected int MarginLeft;
  protected int MarginTop;
  protected int MarginRight;
  protected int MarginBottom;

  protected int PaddingLeft;
  protected int PaddingRight;
  protected int PaddingTop;
  protected int PaddingBottom;

  // Bound coordinates are absolute (relative to window origin), only to change by
  // containers when rendering
  protected Bounds Bounds = new Bounds(0, 0, 0, 0);

  /*
   * Visuals
   */
  protected String Text = ""; // multi-purpose text to display, e.g. button text, label text, textbox content
  protected int FontSize;
  protected int TextAlign = 3; // options can be set with the static fields LEFT, CENTER, RIGHT from PConstants
  // (= 37, 3, 39);

  protected int Background = COLOR; // mode to fill background, default Color
  public static final int COLOR = 0; // use color to draw background
  public static final int IMAGE = 1; // use image to draw background

  protected PImage Image;
  protected int ImageMode = FILL;
  // IMAGE MODES
  public static final int FILL = 1; // distort image if necessary
  public static final int FIT = 2; // enlarge image so it fills out entire object centered, parts of the image
  // might behidden; fits the larger side
  public static final int FITINSIDE = 3; // place image centered in the object without cropping it, so it fits the
  // smaller side

  protected int BackgroundColor;
  protected int ForegroundColor;
  protected int BorderColor;
  protected int HoverColor; // first set automatically with backgroundColor, can be changed programatically
  protected int PressedColor; // first set automatically with backgroundColor, can be changed programatically

  protected int VisualBackgroundColor; // actually displayed color, usually = BackgroundColor, except when hover or

  protected int BorderWidth;
  protected int BorderRadius;
  protected int Cursor;

  /*
   * Mouse Events
   */

  // previous hovered/pressed state
  protected boolean pHovered = false;
  protected boolean pPressed = false;

  // if true then shortcuts registered to frame won't be handled if this element
  // is focused. I.e. useful for textboxes (ctrl-c, x etc.)
  protected boolean overridesParentsShortcuts = false;

  public Control(PApplet pa) {
    papplet = pa;

    Height = 50;
    Width = 50;

    BackgroundColor = 255;
    VisualBackgroundColor = 255;
    ForegroundColor = 0;
    BorderColor = 20;
    HoverColor = 200;
    PressedColor = 150;

    FontSize = 10;

    registeredListeners = new boolean[numberOfListeners];
    registeredMethods = new Method[numberOfListeners];
    registeredMethodWithArgs = new boolean[numberOfListeners];
  }

  /*
   * DRAWING AND RENDERING
   */

  // main drawing method, must be specified for each control individually
  protected abstract void render();

  // just return the looks of this control, without drawing
  protected PImage getGraphics() {
    return pg;
  }

  // call parent to update and indicate that this control has changed its looks
  protected void update() {
    if (frame != null && parent != null) {
      changedVisuals = true;
      parent.update();
    }
  }

  // enable the programmer to force a re-render. This shouldn't be needed, but
  // just in case
  public void forceRenderAlthoughItShouldntBeNeeded() {
    update();
  }

  // specify the overall parent frame (Master GUI Container). This is initiated by
  // the frame itself
  protected void setParentFrame(Frame f) {
    this.frame = f;
  }

  // request frame to focus/blur this object
  public void Focus() {
    frame.requestFocus(this);
  }

  public void Blur() {
    frame.requestBlur(this);
  }

  // just a useful debugging function
  protected void drawBounds() {
    papplet.rect(Bounds.X0, Bounds.Y0, Bounds.X - Bounds.X0, Bounds.Y - Bounds.Y0);
  }

  // standard text drawing method regarding padding and TextAlign
  // this method can be used by any control to its text
  protected void drawDefaultText() {
    pg.fill(ForegroundColor);
    pg.textSize(FontSize);
    switch (TextAlign) {
    case 37: // LEFT
      pg.textAlign(37, 3);
      pg.text(Text, PaddingLeft, (Height - PaddingTop - PaddingBottom) / 2 + PaddingTop);
      break;
    case 3: // CENTER
      pg.textAlign(3, 3);
      pg.text(Text, (Width - PaddingRight - PaddingLeft) / 2 + PaddingLeft, 
        (Height - PaddingTop - PaddingBottom) / 2 + PaddingTop);
      break;
    case 39: // RIGHT
      pg.textAlign(39, 3);
      pg.text(Text, Width - PaddingRight, (Height - PaddingTop - PaddingBottom) / 2 + PaddingTop);
      break;
    }
  }

  // standard background drawing method - this method can be used by any control
  // to its text
  protected void drawDefaultBackground() {
    if (Background == COLOR) {

      if (VisualBackgroundColor != 0) {
        pg.fill(VisualBackgroundColor);
        pg.strokeWeight(BorderWidth);
        pg.stroke(BorderWidth != 0 ? BorderColor : VisualBackgroundColor); // ensure that there's no nasty light
        // border when borderwidth == 0
        pg.rect(BorderWidth / 2, BorderWidth / 2, Width - BorderWidth, Height - BorderWidth, BorderRadius);
      }
    } else if (Background == IMAGE) {

      if (ImageMode == FILL) {
        pg.image(Image, 0, 0, Width, Height);
      } else if (ImageMode == FIT) {
        if (Image.width / Image.height < Width / Height) {
          int newHeight = (int) (Image.height / (float) Image.width * Width);
          pg.image(Image, 0, -(newHeight - Height) / 2, Width, newHeight);
        } else {
          int newWidth = (int) (Image.width / (float) Image.height * Height);
          pg.image(Image, -(newWidth - Width) / 2, 0, newWidth, Height);
        }
      } else if (ImageMode == FITINSIDE) {
        if (Image.width / Image.height > Width / Height) {
          int newHeight = (int) (Image.height / (float) Image.width * Width);
          pg.image(Image, 0, -(newHeight - Height) / 2, Width, newHeight);
        } else {
          int newWidth = (int) (Image.width / (float) Image.height * Height);
          pg.image(Image, -(newWidth - Width) / 2, 0, newWidth, Height);
        }
      }

      // still draw border
      if (BorderWidth > 0) {
        pg.noFill();
        pg.strokeWeight(BorderWidth);
        pg.stroke(BorderWidth != 0 ? BorderColor : VisualBackgroundColor); // ensure that there's no nasty light
        // border when borderwidth == 0
        pg.rect(BorderWidth / 2, BorderWidth / 2, Width - BorderWidth, Height - BorderWidth, BorderRadius);
      }
    }
  }

  /*
   * Style Setter
   */
  public void setX(int x) {
    this.X = x;
    update();
  }

  public void setY(int y) {
    this.Y = y;
    update();
  }

  public void setZ(int z) {
    this.Z = z;
    update();
  }

  public void setPosition(int x, int y) {
    this.X = x;
    this.Y = y;
    update();
  }

  public void setWidth(int Width) {
    this.Width = Width;
    update();
  }

  public void setMinWidth(int Width) {
    this.MinWidth = Width;
    autosize();
    update();
  }

  public void setMaxWidth(int Width) {
    this.MaxWidth = Width;
    autosize();
    update();
  }

  public void setHeight(int Height) {
    this.Height = Height;
    update();
  }

  public void setMinHeight(int Height) {
    this.MinHeight = Height;
    autosize();
    update();
  }

  public void setMaxHeight(int Height) {
    this.MaxHeight = Height;
    autosize();
    update();
  }

  public void setSize(int Width, int Height) {
    this.Width = Width;
    this.Height = Height;
    update();
  }

  public void setBackgroundColor(int clr) {
    BackgroundColor = clr;
    VisualBackgroundColor = clr;
    Background = Control.COLOR;
    update();
  }

  // automatically generates hover and pressed color
  protected void setBackgroundColorsAutomatically(int clr) {
    BackgroundColor = clr;
    VisualBackgroundColor = clr;

    int r = (int) papplet.red(clr);
    int g = (int) papplet.green(clr);
    int b = (int) papplet.blue(clr);
    int a = (int) papplet.alpha(clr);

    if (papplet.brightness(clr) > 40) { // darken color for HoverColor and PressedColor when color is bright enough
      HoverColor = Color.colors(r - 20, g - 20, b - 20);
      PressedColor = Color.colors(r - 40, g - 40, b - 40);
    } else {// lighten color for HoverColor and PressedColor when color too dark
      HoverColor = Color.colors(r + 20, g + 20, b + 20);
      PressedColor = Color.colors(r + 40, g + 40, b + 40);
    }
    Background = Control.COLOR;
    update();
  }

  public void setForegroundColor(int clr) {
    ForegroundColor = clr;
    update();
  }

  public void setHoverColor(int clr) {
    this.HoverColor = clr;
    update();
  }

  public void setPressedColor(int clr) {
    this.PressedColor = clr;
    update();
  }

  public void setBorderColor(int clr) {
    BorderColor = clr;
    update();
  }

  public void setBorderWidth(int w) {
    this.BorderWidth = Math.max(0, w);
    update();
  }

  public void setBorderRadius(int r) {
    this.BorderRadius = r;
  }

  public void setText(String txt) {
    this.Text = txt;
    autosize();
    update();
  }

  public void setFontSize(int s) {
    this.FontSize = s;
    autosize();
    update();
  }

  public void setTextAlign(int align) {
    this.TextAlign = align;
  }

  public void setCursor(int c) {
    if (c >= 0 && c < 12) {
      Cursor = c;
    }
  }

  public void setImage(PImage i) {
    try {
      this.Image = (PImage) i.clone();
      Background = Control.IMAGE;
    } 
    catch (Exception e) {
      e.printStackTrace();
    }
  }

  public void setImageMode(int mode) {
    ImageMode = mode;
  }

  public void setGradient(int clr1, int clr2) {
    PGraphics gradient = papplet.createGraphics(Width, Height);
    gradient.beginDraw();
    for (int i = 0; i < Height; i++) {
      float inter = PApplet.map(i, 0, Height, 0, 1);
      int c = papplet.lerpColor(clr1, clr2, inter);
      gradient.stroke(c);
      gradient.line(0, i, Width, i);
    }
    setImage(gradient);
  }

  public void setEnabled(boolean state) {
    this.Enabled = state;
    update();
  }

  public void setVisible(boolean state) {
    this.Visible = state;
    update();
  }

  public void setMargin(int all) {
    MarginTop = all;
    MarginRight = all;
    MarginBottom = all;
    MarginLeft = all;
    update();
  }

  public void setMargin(int topBottom, int leftRight) {
    MarginTop = topBottom;
    MarginRight = leftRight;
    MarginBottom = topBottom;
    MarginLeft = leftRight;
    update();
  }

  public void setMargin(int top, int right, int bottom, int left) {
    MarginTop = top;
    MarginRight = right;
    MarginBottom = bottom;
    MarginLeft = left;
    update();
  }

  public void setMarginTop(int top) {
    MarginTop = top;
    update();
  }

  public void setMarginRight(int right) {
    MarginRight = right;
    update();
  }

  public void setMarginBottom(int bottom) {
    MarginBottom = bottom;
    update();
  }

  public void setMarginLeft(int left) {
    MarginLeft = left;
    update();
  }

  public void setPadding(int all) {
    PaddingTop = all;
    PaddingRight = all;
    PaddingBottom = all;
    PaddingLeft = all;
    autosize();
    update();
  }

  /*
   * Autosize specifies actions that will set width/height, when padding, Text, and FontSize changed
   * Each class can override it to specify custom calculations
   */
  protected void autosize() {
  }

  public void setPadding(int topBottom, int leftRight) {
    PaddingTop = topBottom;
    PaddingRight = leftRight;
    PaddingBottom = topBottom;
    PaddingLeft = leftRight;
    autosize();
    update();
  }

  public void setPadding(int top, int right, int bottom, int left) {
    PaddingTop = top;
    PaddingRight = right;
    PaddingBottom = bottom;
    PaddingLeft = left;
    autosize();
    update();
  }

  public void setPaddingTop(int top) {
    PaddingTop = top;
    autosize();
    update();
  }

  public void setPaddingRight(int right) {
    PaddingRight = right;
    autosize();
    update();
  }

  public void setPaddingBottom(int bottom) {
    PaddingBottom = bottom;
    autosize();
    update();
  }

  public void setPaddingLeft(int left) {
    PaddingLeft = left;
    autosize();
    update();
  }

  /*
   * STYLE GETTER
   */

  public int getX() {
    return X;
  }

  public int getY() {
    return Y;
  }

  public int getWidth() {
    return Width;
  }

  public int getMinWidth() {
    return MinWidth;
  }

  public int getMaxWidth() {
    return MaxWidth;
  }

  public int getHeight() {
    return Height;
  }

  public int getMinHeight() {
    return MinHeight;
  }

  public int getMaxHeight() {
    return MaxHeight;
  }

  public int getBackgroundColor() {
    return BackgroundColor;
  }

  public int getForegroundColor() {
    return ForegroundColor;
  }

  public int getHoverColor() {
    return HoverColor;
  }

  public int getPressedColor() {
    return PressedColor;
  }

  public int getBorderColor() {
    return BorderColor;
  }

  public int getBorderWidth() {
    return BorderWidth;
  }

  public int getBorderRadius() {
    return BorderRadius;
  }

  public String getText() {
    return Text;
  }

  public int getFontSize() {
    return FontSize;
  }

  public String getTextAlign() {
    switch (TextAlign) {
    case 37:
      return "LEFT";
    case 3:
      return "CENTER";
    case 39:
      return "RIGHT";
    default:
      return "";
    }
  }

  public int getCursor() {
    return Cursor;
  }

  public PImage getImage() {
    return Image;
  }

  public boolean getEnabled() {
    return Enabled;
  }

  public boolean getVisible() {
    return Visible;
  }

  public boolean isFocusable() {
    return Focusable;
  }

  public boolean getFocused() {
    return Focused;
  }

  public int getMarginTop() {
    return MarginTop;
  }

  public int getMarginRight() {
    return MarginRight;
  }

  public int getMarginBottom() {
    return MarginBottom;
  }

  public int getMarginLeft() {
    return MarginLeft;
  }

  public int getPaddingTop() {
    return PaddingTop;
  }

  public int getPaddingRight() {
    return PaddingRight;
  }

  public int getPaddingBottom() {
    return PaddingBottom;
  }

  public int getPaddingLeft() {
    return PaddingLeft;
  }

  /*
   * EVENT METHODS
   */

  // true when listener has been set
  protected int numberOfListeners = 5; // for basic control 5, but some control children use more
  protected boolean registeredListeners[]; // press, release, enter, exit, wheel
  protected boolean registeredMethodWithArgs[]; // press, release, enter, exit, wheel
  protected Method registeredMethods[];

  public boolean addEventListener(String type, String methodName) {
    if (!registeredMouseListener)
      addMouseListener(); // needed for controls that don't register a listener by default;
    switch (type) {
    case "press":
      return registerEventMethod(0, methodName);
    case "release":
      return registerEventMethod(1, methodName);
    case "enter":
      return registerEventMethod(2, methodName);
    case "exit":
      return registerEventMethod(3, methodName);
    case "wheel":
      return registerEventMethod(4, methodName);
    default:
      return false;
    }
  }

  public void removeEventListener(String type) {
    switch (type) {
    case "press":
      deregisterEventMethod(0);
      break;
    case "release":
      deregisterEventMethod(1);
      break;
    case "enter":
      deregisterEventMethod(2);
      break;
    case "exit":
      deregisterEventMethod(3);
      break;
    case "wheel":
      deregisterEventMethod(4);
      break;
    default:
    }
  }

  protected boolean registerEventMethod(int number, String methodName) {
    if (!registeredListeners[number]) { // not yet registered a method for this event
      Class<?> c = papplet.getClass();
      try { // try with args
        registeredMethods[number] = c.getMethod(methodName, MouseEvent.class);
        registeredListeners[number] = true;
        registeredMethodWithArgs[number] = true;
        return true;
      } 
      catch (NoSuchMethodException nsme) {
        try { // try with no args
          registeredMethods[number] = c.getMethod(methodName);
          registeredListeners[number] = true;
          registeredMethodWithArgs[number] = false;
          return true;
        } 
        catch (NoSuchMethodException nsme2) {
          papplet.die("There is no public " + methodName + "() method.");
          registeredListeners[number] = false;
        }
      }
    }
    return false;
  }

  protected void deregisterEventMethod(int number) {
    registeredMethods[number] = null;
    registeredListeners[number] = false;
    registeredMethodWithArgs[number] = false;
  }

  protected boolean registeredMouseListener = false;

  protected void addMouseListener() {
    if (!registeredMouseListener) {
      // papplet.registerMethod("mouseEvent", this);
      registeredMouseListener = true;
    }
  }

  protected void removeMouseListener() {
    if (registeredMouseListener) {
      papplet.unregisterMethod("mouseEvent", this);
      registeredMouseListener = false;
    }
  }

  // has to be public
  public void mouseEvent(MouseEvent e) {
    int x = e.getX();
    int y = e.getY();

    if (Visible && Enabled) {

      if (x > Bounds.X0 && x < Bounds.X && y > Bounds.Y0 && y < Bounds.Y) { // if over element

        switch (e.getAction()) {
          case MouseEvent.PRESS:
            press(e);
          handleRegisteredEventMethod(0, e);
          break;
        case MouseEvent.RELEASE:
          release(e);
          handleRegisteredEventMethod(1, e);
          break;
        case MouseEvent.DRAG:
          drag(e);
          break;
        case MouseEvent.WHEEL:
          mouseWheel(e);
          handleRegisteredEventMethod(4, e);
          break;
        }
        if (!pHovered) { // ENTER
          papplet.cursor(Cursor);
          enter(e);
          handleRegisteredEventMethod(2, e);
          pHovered = true;
        }
      } else {
        if (pHovered) { // EXIT
          papplet.cursor(0);
          exit(e);
          handleRegisteredEventMethod(3, e);
          pHovered = false;
        }
      }
    }
  }

  protected void handleRegisteredEventMethod(int number, MouseEvent e) {
    if (registeredListeners[number]) {
      try {
        if (registeredMethodWithArgs[number] == true)
          registeredMethods[number].invoke(papplet, e);
        else
          registeredMethods[number].invoke(papplet);
      } 
      catch (IllegalAccessException ie) {
        ie.printStackTrace();
      } 
      catch (InvocationTargetException te) {
        te.printStackTrace();
      }
    }
  }

  protected void press(MouseEvent e) {
  }

  protected void release(MouseEvent e) {
  }

  protected void enter(MouseEvent e) {
  }

  protected void exit(MouseEvent e) {
  }

  protected void mouseWheel(MouseEvent e) {
  }

  protected void drag(MouseEvent e) {
  }

  // methods that will be called when key events occured and this Control is the
  // currently focused element.
  public void onKeyPress(KeyEvent e) {
  }

  public void onKeyRelease(KeyEvent e) {
  }

  public void onKeyTyped(KeyEvent e) {
  }
}
