package pGUI.core;

import processing.core.*;
import processing.event.*;
import pGUI.classes.*;

import java.lang.Exception;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.ArrayList;

/*
 * Frame is the main container for all controls in one sketch and there must always be one frame per project. 
 * It handles all rendering of the elements, updating looks and feels and finally displaying the GUI. 
 * 
 * It is itself a container that fills out the entire size of the sketch window by default. 
 * 
 * When creating the frame you have to pass THE PARENT PAPPLET TO THE CONTRUCTOR in setup() like this: Frame f = new Frame(this);
 * 
 */

public class Frame extends Container {

	// the element that has requested focus last.
	// this element will always get the keyEvents going on.
	protected Control FocusedElement = this;
	protected boolean initialized = false;

	protected final int PRE = 1; // draw GUI before draw() takes place
	protected final int POST = 2; // draw GUI after draw() takes place
	protected int timeToDraw = PRE;

	protected float Scale = 1;

	protected boolean chokeMouseListeners = false;

	public Frame(PApplet pa) {
		super(pa);
		papplet = pa;
		frame = this;

		papplet.registerMethod("pre", this);

		// default size fills out entire sketch window
		setSize(papplet.width, papplet.height);
		Bounds.X = papplet.width;
		Bounds.Y = papplet.height;

		addKeyEventListener();

		animations = new ArrayList<Animation>();

		papplet.registerMethod("mouseEvent", this);
	}

	public Frame(PApplet pa, int timeToDraw) { // option with determining when the gui should be draw - pre (1) as
												// standard and post (2) when the user wants the gui to overdraw the
												// window
		super(pa);
		papplet = pa;
		frame = this;

		if (timeToDraw == PRE) {
			papplet.registerMethod("pre", this);
		} else if (timeToDraw == POST) {
			papplet.registerMethod("draw", this);
		}

		Bounds.X = papplet.width;
		Bounds.Y = papplet.height;

		// default size fills out entire sketch window
		setSize(papplet.width, papplet.height);
		addKeyEventListener();

		animations = new ArrayList<Animation>();
	}

	// before draw() already draw GUI, so the user may draw above it, has t be
	// public
	public void pre() {
		this.display();
	}

	public void draw() {
		this.display();
	}

	public void display() {

		// int t0 = millis();
		if (!initialized) {
			setParentFrame();
			initialized = true;

			// sort by z-Index, from lowest to highest
			sortContent();

		}

		// handle animations
		for (int i = 0; i < animations.size(); i++) {
			if (!animations.get(i).animate()) { // if animation returns false it has finished and can be removed
				animations.remove(i);
			}
		}

		if (changedVisuals && Visible) {
			changedVisuals = false;
			render();
		}

		// scale(Scale);
		papplet.image(pg, 0, 0);
		// papplet.set(0, 0, pg); //ist wirklich am schnellsten
		/*
		 * papplet.loadPixels(); for(int i = 0; i < pg.pixels.length;i++){
		 * papplet.pixels[i] = pg.pixels[i]; } papplet.updatePixels();
		 */
		// println(millis()-t0);
		chokeMouseListeners = false;
	}

	@Override // - Frame doesn't need to call its parent to update
	protected void update() {
		changedVisuals = true;
	}

	// call "setParentFrame" for all children. If child is a container, it will set
	// it's own parentFrame and call its children setParentFrame()

	public void setParentFrame() {
		for (int i = 0; i < Content.size(); i++) {
			Content.get(i).setParentFrame(this);
		}
	}

	/*
	 * SHORTCUTS
	 */

	class ShortcutDetails { // store a method and the object it shall be invoked on in one class
		Method m;
		Object o;
		boolean global; // if shortcut is global then it will be executed even if the focused element
						// ovverrides shortcuts

		ShortcutDetails(Method m, Object o, boolean global) {
			this.m = m;
			this.o = o;
			this.global = global;
		}
	}

	// list of registered shortcuts
	protected HashMap<Shortcut, ShortcutDetails> ShortcutMethods = new HashMap<Shortcut, ShortcutDetails>();

	public boolean registerShortcut(Shortcut s, String methodName, Object target) {
		return registerShortcut(s, methodName, target, false);
	}

	public boolean registerShortcut(Shortcut s, String methodName, Object target, boolean global) {
		Class<?> c = target.getClass();
		try {

			Method method = c.getMethod(methodName);
			ShortcutMethods.put(s, new ShortcutDetails(method, target, global));
		} catch (NoSuchMethodException nsme) {
			papplet.die("There is no public " + methodName + "() method in the class " + target.getClass().getName());
		} catch (Exception e) {
			papplet.die("Could not register " + methodName + " + () for " + target, e);
		}
		return false;
	}

	protected boolean checkShortcut(Shortcut s) {

		ShortcutDetails o = ShortcutMethods.get(s);
		if (o != null) {
			if (!FocusedElement.overridesParentsShortcuts || o.global) { // don't handle shortcut if focused element
																			// overrides shortcuts, but only if shortcut
																			// isn't global
				handleShortcut(o);
			}
			return true;
		}
		return false;
	}

	protected void handleShortcut(ShortcutDetails om) {
		Method method = om.m;
		Object object = om.o;
		try {
			method.invoke(object);
		} catch (Exception e) {
			// check for wrapped exception, get root exception
			Throwable t;
			/*
			 * if (e instanceof InvocationTargetException) { InvocationTargetException ite =
			 * (InvocationTargetException) e; t = ite.getCause(); } else
			 */ {
				t = e;
			}
			// check for RuntimeException, and allow to bubble up
			if (t instanceof RuntimeException) {
				// re-throw exception
				throw (RuntimeException) t;
			} else {
				// trap and print as usual
				t.printStackTrace();
			}
		}
	}

	/*
	 * FOCUS HANDLING
	 * 
	 * when other controls request focus or blur
	 * 
	 */

	protected boolean requestFocus(Control c) {
		if (c.Focusable && !FocusedElement.stickyFocus) {
			FocusedElement.Focused = false;
			FocusedElement.update();
			FocusedElement = c;
			FocusedElement.Focused = true;
			return true;
		}
		return false;
	}

	protected void requestBlur(Control c) {
		if (FocusedElement == c) {
			FocusedElement.Focused = false;
			FocusedElement.update();
			FocusedElement = this;
		}
	}

	/*
	 * ANIMATIONS
	 *
	 */

	ArrayList<Animation> animations;

	public void animate(String attribute, Object target, int aimedValue, double increment) {

		Animation newAnimation = new Animation(attribute, target, aimedValue, increment);

		// replace animation of same kind (same object and attribute)
		for (int i = 0; i < animations.size(); i++) {
			if (animations.get(i).compare(newAnimation)) {
				animations.remove(i);
			}
		}
		animations.add(newAnimation);
	}

	/*
	 * WINDOW SETTER
	 */

	/*
	 * KEY EVENTS
	 */

	// there's only one keyListener per GUI, it hands the recognized keystrokes over
	// to the FocusedElement
	// Independently of the focusedElement keystrokes are always compared to the
	// registered shortcuts.

	protected KeyListener keyListener;

	protected void addKeyEventListener() {
		if (keyListener == null) {
			papplet.registerMethod("keyEvent", this);
			keyListener = new KeyListener(this);
		}
	}

	// has to be public
	public void keyEvent(KeyEvent e) {
		keyListener.handleKeyEvent(e, FocusedElement);
	}
}