
/*
 * ECScreen is a class for displaying menus, letters and symbols in a way the user
 * can easily interpret as iContact-language commands. 
 *
 * .txt-files that are in a special EScreen format (xml-oriented) can be read
 *    by the special parser. 
 * These files require:
 *    - a wrapper: <ECScreen></ECScreen>
 *    - and a <body></body> within that
 * you can give information about the screen in <title></title> and <description></description>
 
 * in the <head></head> you can specify text alignment with the keywords LEFT, CENTER, RIGHT and TOP, MIDDLE, BOTTOM ...
 * ... and the so-called enterMode. It will give the program information about how eye movement information shall be interpreted
 *   Options are: 
 *      enterMode="select-press" - for navigation and simple screens. You look to one of the 9 directions and select by activating the extra sensor
 *      enterMode="two-way"      - for large-scaled screens that feature a lot of elements. The user looks in two directions (shouldn't be center) and the result action is immediate
 
 
 * In the <body></body> you can add elements. Currently only letters (strings) and lines are supported
 *    <letters>
 *        a: 14,14
 *        b: 23,3
 *    </letters>
 *          first you give the string followed by a double stop and the x- and y-coordinate, seperated by comma
 
 *    <lines>
 *        x0,x1,y0,y1
 *    </lines>
 *          - pretty obvious
 
 
 * You can also set text size and color in the constructor or with setFontSize(int), setFontSize(int)
 
 * IT IS IMPORTANT TO ADD "ECScreenConstants.drawECScreen" (for currently active screen) or "SCREEN.drawECScreen();" (for specific screen) TO THE draw() FUNCTION!!
 *
 */






static class ECScreenConstants {
  private  static ECScreen activeScreen; //currently displayed ECScreen
  private static ArrayList<ECScreen> screens = new ArrayList<ECScreen>(); //array of all avtive and available ECScreens

  public static boolean setActiveScreen(int s) { //set displayed screen via id
    boolean successful = false;
    for (int i = 0; i < screens.size(); i++) {
      if (s == screens.get(i).getId()) {
        activeScreen = screens.get(i);
        break;
      }
    }
    return successful;
  }
  public static boolean setActiveScreen(String s) { //set displayed screen via title
    boolean successful = false;
    for (int i = 0; i < screens.size(); i++) {
      if (s.equals(screens.get(i).getTitle())) {
        activeScreen = screens.get(i);
        break;
      }
    }
    return successful;
  }
  public static void drawScreen() { //main methods, should be called in main each frame
    if (activeScreen != null) {
      activeScreen.drawECScreen();
    }
  }

  public static int getActiveScreenId() { //get id of the active screen
    if (activeScreen != null) {
      return activeScreen.getId();
    } else {
      return -1;
    }
  }
}






class ECScreen {
  private int id;                         //can be used to identify ECScreens, should start with 1. individual numbers recommended!
  private String fileName;                //stores the name of the file that was used to get the data

  private String title;
  private String description;
  private int textHAlign = LEFT;          //horizontal text align
  private int textVAlign = BASELINE;      //vertical text align
  private int enterMode = 2;              //1: select-press, 2: two-way:
  private int enterContent = 2;           //1: functions, 2: string

  private int numberLetters;              //number of elements featured on the screen
  private String letters[];               //elements that appear on the screen
  private float lettersPos[][];           //position for every element in percent. first dimension refers to element (parallel to elements[]), second to x (0), and y (1)-coordinate
  private int numberLines;
  private int lines[][];
  private int fontSize;
  private color fontColor;

  private int h;                          //height of ScreenPlane (usually = PApplet height
  private int w;                          //width of ScreenPlane
  private PGraphics ScreenPlane;          //PGraphic on which the screen is drawn only once(!) and then only projected onto surface for performance

  private boolean initialized = false;    //is set to true when parsing was succesful

  public boolean visible = true;


  ECScreen(String fileName, int id, int fontSize) {
    this.id = id;
    this.fileName = fileName;
    this.fontSize = fontSize;
    this.fontColor = 0; //standard color in case not set in file
    this.w = width;
    this.h = height;
    String[] lines = loadStrings(fileName);

    if (lines != null) {
      if (this.parseFile(lines)) {
        this.initialized = true;
        this.createGraphic();
      }
      ECScreenConstants.screens.add(this);
    } else {
      println("error. File not found");
    }
  }
  ECScreen(String fileName, int id, int fontSize, color fontColor) { 
    this.id = id;
    this.fileName = fileName;
    this.fontSize = fontSize;
    this.w = width;
    this.h = height;
    String[] lines = loadStrings(fileName);

    if (lines != null) {
      if (this.parseFile(lines)) {
        this.initialized = true;
        this.fontColor = fontColor; //overwrites color settings from file
        this.createGraphic();
      }
      ECScreenConstants.screens.add(this);
    } else {
      println("error. File not found");
    }
  }






  private boolean parseFile(String[] fileLines) { //parse file and store the values
    String txt = join(fileLines, "\n");

    //check if file is ECScreen format
    String ecs = parseSection("<ECScreen>", txt);
    if (ecs != null) {

      //get title of screen
      String title = parseSection("<title>", txt);
      this.title = (title != null) ? title : "no title";

      //get description of ECScreen
      String desc = parseSection("<description>", txt);
      this.description = (desc != null) ? desc : "no description";


      //head, get infos about it (number of letters/words and lines), not used at the time
      String head = parseSection("<head>", txt);
      if (head != null) {
        if (head.indexOf("LEFT") > -1) { 
          textHAlign = LEFT;
        } else if (head.indexOf("CENTER") > -1) {
          textHAlign = CENTER;
        } else if (head.indexOf("RIGHT") > -1) {
          textHAlign = RIGHT;
        }
        if (head.indexOf("TOP") > -1) {
          textVAlign = TOP;
        } else if (head.indexOf("BOTTOM") > -1) {
          textVAlign = BASELINE;
        } else if (head.indexOf("MIDDLE") > -1) {
          textVAlign = CENTER;
        }
        String[] enterModeOcurrences = match(head, "(?<=enterMode=\")([a-z]|[A-Z]|-)+(?=\")");
        if (enterModeOcurrences != null) {
          if (enterModeOcurrences[0].equals("select-press")) {
            enterMode = 1;
          } else if (enterModeOcurrences[0].equals("two-way")) {
            enterMode = 2;
          }
        }
        String[] enterContentOccurrences = match(head, "(?<=enterContent=\")([a-z]|[A-Z]|-)+(?=\")");
        if (enterContentOccurrences != null) {
          if (enterContentOccurrences[0].equals("functions")) {
            enterContent = 1;
          } else if (enterContentOccurrences[0].equals("string")) {
            enterContent = 2;
          }
        }
        String[] textColors = match(head, "(?<=color=\")(\\d+, ?\\d+, ?\\d+)+(?=\")");
        if (textColors != null) {
          String[] rgbS = split(textColors[0], ',');
          int[] rgbI = new int[3];
          rgbI[0] = int(trim(rgbS[0]));
          rgbI[1] = int(trim(rgbS[1]));
          rgbI[2] = int(trim(rgbS[2]));
          color clr = color(rgbI[0], rgbI[1], rgbI[2]);
          this.fontColor = clr;
        }
      }

      //body, parse information, error when no body!
      String body = parseSection("<body>", txt);
      if (body != null) {
        //PARSE LETTERS
        String letters = parseSection("<letters>", body);
        if (letters != null) {
          String[] rawLetterContent = splitTokens(letters, "\n");
          StringList letterContent = new StringList();
          for (int i = 0; i < rawLetterContent.length; i++) {
            if (rawLetterContent[i].indexOf("//") == -1 && trim(rawLetterContent[i])!="") {
              letterContent.append(rawLetterContent[i]);
            }
          }
          this.numberLetters = letterContent.size();
          this.letters = new String[this.numberLetters];
          this.lettersPos = new float[this.numberLetters][2];

          for (int i = 0; i < letterContent.size(); i++) { //parse all letters individually
            String[] letter = trim(match(letterContent.get(i), ".+:"));
            if (letter != null) {
              this.letters[i] = trim(letter[0].substring(0, letter[0].length()-1));
            }
            String[] posX = match(letterContent.get(i), "[0-9]+,");
            int positionX = 0;
            if (posX != null) {
              positionX = int(posX[0].substring(0, posX[0].length()-1));
            }
            String[] posY = match(letterContent.get(i), ", ?[0-9]+");
            int positionY = 0;
            if (posY != null) {
              positionY = int(trim(posY[0].substring(1)));
            }
            this.lettersPos[i][0] = positionX;
            this.lettersPos[i][1] = positionY;
          }
        }
        //PARSE LINES
        String lines = parseSection("<lines>", body);
        if (lines != null) {
          String[] linesContent = splitTokens(lines, "\n");
          this.numberLines = linesContent.length;
          this.lines = new int[this.numberLines][4];
          for (int i = 0; i < linesContent.length; i++) {
            String row[] = splitTokens(linesContent[i], ",");
            this.lines[i][0] = int(trim(row[0]));
            this.lines[i][1] = int(trim(row[1]));
            this.lines[i][2] = int(trim(row[2]));
            this.lines[i][3] = int(trim(row[3]));
          }
        }
        String interpreter = parseSection("<interpreter>", body);
        if (interpreter != null) {
          interpreterParser(interpreter);
        }
        System.out.print("-Loaded ECScreen \"" + title + "\"  from \'" + fileName + "\'");
        System.out.print(" (" + ((enterContent==1)?"function":"input") + " screen)\n"); //print enterContent type
        println(" Description: " + description);
        return true;
      } else {
        println("Error with ECScreen parsing: no body!");
        return false;
      }
    } else {
      println("Error with ECScreen parsing: no ECScreen format");
      return false;
    }
  }




  //parse a section in xml style, with <></>-coding
  //searches for String command (i.e. <body>) and returns the content of the first occurence 
  private String parseSection(String command, String txt) {
    String[] section = match(txt, command + ".*</" + command.substring(1));
    if (section != null) {
      String sectionContent = trim(section[0].substring(command.length(), section[0].length()-command.length()-1));
      return sectionContent;
    } else {
      return null;
    }
  }




  private void interpreterParser(String txt) {
    //first get dimensions
    int w = 9, h=9;
    String[] iWidth = match(txt, "(?<=width=\")[0-9](?=\")");
    if (iWidth != null) {
      w = int(iWidth[0]);
    }
    String[] iHeight = match(txt, "(?<=height=\")[0-9](?=\")");
    if (iHeight != null) {
      h = int(iHeight[0]);
    }
  }





  public void setFontSize(int fontSize) {
    this.fontSize = fontSize;
  }

  public int getFontSize() {
    return fontSize;
  }

  public void setFontColor(color clr) {
    fontColor = clr;
  }

  public color getFontColor() {
    return fontColor;
  }





  private void createGraphic() {
    ScreenPlane = createGraphics(w, h);
    ScreenPlane.beginDraw();
    ScreenPlane.stroke(0);
    ScreenPlane.fill(fontColor);
    ScreenPlane.textSize(fontSize);
    ScreenPlane.textAlign(textHAlign, textVAlign);
    ScreenPlane.strokeWeight(2);
    for (int i = 0; i < numberLetters; i++) {
      ScreenPlane.text(letters[i], lettersPos[i][0] * w / 100, lettersPos[i][1] * h / 100);
    }
    for (int i = 0; i < numberLines; i++) {
      ScreenPlane.line(lines[i][0] * w / 100, lines[i][2] * h / 100, lines[i][1] * w / 100, lines[i][3] * h / 100);
    }
    ScreenPlane.endDraw();
  }



  public void drawECScreen() { //draw Screen on PApplet, Graphics aren't rendered new. Use redraw instead
    if (initialized) {
      if (visible) {
        image(ScreenPlane, (w - ScreenPlane.width) / 2, (h - ScreenPlane.height) / 2);
      }
    } else {
      println("Exception: Screen hasn't been parsed/intialized properly");
    }
  }



  public void redraw() { //re-render the graphics, when size or color has been changed
    createGraphic();
  }



  public void print() { //print list of letters and lines
    for (int i = 0; i < letters.length; i++) {
      println(letters[i] + "\t" + lettersPos[i][0] + "\t" + lettersPos[i][1]);
    }
    for (int i = 0; i < numberLines; i++) {
      println(lines[i][0], lines[i][1], lines[i][2], lines[i][3]);
    }
  }



  public int getId() {
    return id;
  }
  public String getTitle() {
    return title;
  }
  public int getActiveScreen() {
    return ECScreenConstants.getActiveScreenId();
  }
}
