/*  //<>// //<>// //<>// //<>//
 * Class for integrating the arduino into the program. 
 * A usb port can be specified and the data from the serial will be read. 
 * UPDATE: Default will be:
 *                   - check for available usb-ports automatically. 
 *                   - then check if the devices sends the string "FLXSNSR" which our arduino does
 *                   - !! this takes some time because it seems the arduino needs about 1.5 seconds to restart and send the data
 * 
 * The method "portList()" returns a String[] array with all available usb ports. 
 * with "setPort(int)"/"setPort(String portName)" the port can be selected and will be set up automatically
 * setPort() returns false, when setup has not been succcessful
 *
 * It is possible to look again for new usb ports with "updateList()"
 *
 *
 */

import processing.serial.Serial;
//import java.util.Map;

interface sensorClick {
  void singleClick();
  void longClick();
}

public class FlexSensor {

  private boolean initialized = false; //will be set to true, when port is specified

  protected boolean connected = false;
  protected boolean connecting = false;

  private String[] portList;
  private String portName;
  private Serial port;
  private int tolerance = 50;
  private int checkConnectionInterval = 2000; // check of connectin is still available every x ms
  private int readDataCounter = 0;
  private int initialNumberOfPorts = 0;
  private PApplet parent;
  private boolean sleeping = false;

  private boolean useNotifications = true;
  Notifier notifier;

  static final int NULL = 0;
  static final int SHORT = 1;
  static final int LONG = 2;
  int signal = NULL;


  FlexSensor(PApplet pa) {
    parent = pa;
    registerMethod("draw", this);
    notifier = new Notifier(250, 60);
    tryConnect();
  }


  FlexSensor(PApplet pa, Notifier n) {  //use externel notifier
    parent = pa;
    registerMethod("draw", this);
    this.notifier = n;
    tryConnect();
  }


  public void draw() {
    if (connecting) {
      parent.resetMatrix();
      parent.colorMode(RGB, 255);
      parent.fill(40);
      parent.rect(0, 8 * height / 9, width, height / 9);
      parent.textAlign(CENTER, CENTER);
      parent.fill(70);
      parent.textSize(30);
      String dots = "";
      if (frameCount %90 > 60) {
        dots = "...";
      } else if (frameCount %90 > 30) {
        dots = ".. ";
      } else {
        dots = ".  ";
      } 
      parent.text("Establishing connection to Flex-Sensor " + dots, width/2, 17 * height / 18);
    } else {
      readPort(true);
    }
  }


  protected void tryConnect() {
    //do the connecting in a seperate thread because it takes some time
    ConnectArduinoThread connectArduinoThread = new ConnectArduinoThread(this);
    connectArduinoThread.start();
  }



  public String[] portList() {
    return portList;
  }

  public void printPortList() {
    if (portList != null) {
      for (int i = 0; i < portList.length; i++) {
        println(portList[i]);
      }
    }
  }



  void readPort(boolean print) {
    int t0 = millis();
    readDataCounter++;

    //if (initialized) { //ensure port has been set
    signal = NULL;
    if (connected) {
      String input = "";
      
      while (port.available() > 0) {
        char inChar = port.readChar(); 
        input += inChar;
      }
      
      if (input.contains("41")) {
        signal = SHORT;
      } else if (input.contains("42")) {
        signal = LONG;
      } else if (input.contains("43")) {
        if (tracker != null) {
          tracker.resetCalibration2();
        }
      } else if (input.contains("$")) {
        notifier.setNotification(input.substring(2,input.length()-4), 1000);
      }
    }


    //check if port disconnected or connected
    if (readDataCounter > frameRate * checkConnectionInterval / 1000) {
      readDataCounter = 0;
      //check ports again
      portList = Serial.list();



      //if detected less devices
      if (portList.length < initialNumberOfPorts) {
        println("less");

        //check if the disconnected device has been the connected sensor
        boolean removedConnectedDevice = true;
        for (int i = 0; i < portList.length; i++) {
          if (portList[i].equals(portName)) { 

            initialNumberOfPorts = portList.length;

            //if found portName in list everything is alright
            return;
          }
        }
        //if this code runs the sensor has been the disconnected device

        if (useNotifications) notifier.setNotification("Your sensor has been removed", 2000);

        initialNumberOfPorts = portList.length;
        port.stop();
        connected = false;

        //initialized = false;
      } 
      //if detected more connected devices, and no connection has been made yet
      else if (portList.length > initialNumberOfPorts && !connected) {
        initialNumberOfPorts = portList.length;
        println("tryc");
        tryConnect();
      }
    }
    timeKeeper.add(millis()-t0, "Read Arduino");
  }
  void connectionMade() {
    if (useNotifications) notifier.setNotification("Sensor has been connected", 2000);
  }

  //____
  //interface methods:
  //_____
  void setTolerance(int t) {
    if (initialized) {
      tolerance = t;
      port.write("t" + t + "\n");
    }
  }

  int getTolerance() { //doesn't check arduino!!
    return tolerance;
  }

  void wake() {
    if (connected) {
      port.write("1\n");
      sleeping = false;
    }
  }
  void sleep() {
    if (connected) {
      port.write("2\n");
      sleeping = true;
    }
  }
  void recalibrate() {
    if (connected) {
      port.write("3\n");
    }
  }
  void enableLongClick() {
    if (initialized) {
      port.write("4\n");
    }
  }
  void disableLongClick() {
    if (initialized) {
      port.write("5\n");
    }
  }



  @Deprecated
    public boolean setPort(String portName) { //returns true when assignment has been successful
    boolean successful = false;
    for (int i = 0; i < this.portList.length; i++) {
      if (this.portList.equals(portName)) {
        this.portName = portName;

        this.port = new Serial(parent, this.portName, 9600);
        if (this.port != null) {
          successful = true;
        }
        break;
      }
    }
    if (successful) {
      initialized = true;
    } else {
      initialized = false;
    }
    return successful;
  }

  @Deprecated
    public boolean setPort(int position) { //returns true when assignment has been successful
    boolean successful = false;
    if (true) {
      return true;
    }
    if (position >= 0 && position < this.portList.length) {
      this.portName = this.portList[position];
      this.port = new Serial(parent, this.portName, 9600);
      if (this.port != null) {
        successful = true;
      }
    }
    if (successful) {
      initialized = true;
    } else {
      initialized = false;
    }
    return successful;
  }
}







class ConnectArduinoThread extends Thread {
  private boolean active;
  private int time;
  private FlexSensor arduino;


  ConnectArduinoThread(FlexSensor arduino) {
    this.arduino = arduino;
    active = false;
  }


  public void start() {
    if (!active) {
      active = true;
      this.time = time;
      super.start();
    }
  }


  void run() {
    while (true) {
      if (active) {
        execute();
      } else {
        break;
      }
    }
  }


  private void execute() {
    arduino.connecting = true;
    println("tryc2");
    connect();

    arduino.connecting = false;

    quit();
  }
  protected boolean connect() {
    arduino.portList = Serial.list();
    arduino.initialNumberOfPorts = arduino.portList.length;
    for (int i = 0; i < arduino.portList.length; i++) {
      try {
        println(arduino.portList[i]);
        arduino.port = new Serial(arduino.parent, arduino.portList[i], 9600) ;      // Open 1st port in list

        delay(2000);

        //perform handshake - ask if device is a flexsensor arduino
        //port.clear();
        //port.write("2\n");


        String input = ""; 
        //read answer
        while (arduino.port.available() > 0) { 
          char inChar = arduino.port.readChar();
          input += inChar;
          //print(inChar);
        }
        println("tryc3");


        if (input.contains("FLXSNSR")) {
          println("connected arduino");
          arduino.portName = arduino.portList[i];
          arduino.connected = true;
          arduino.connectionMade();
          break;
        } else {
          arduino.port.stop();
        }
      }  
      catch (Exception e) {
        e.printStackTrace();
        arduino.port.stop();
        return false;
      }
    }
    return false;
  }

  public void quit() {
    active = false;
    interrupt();
  }
}
