/*
 * Class for Graphical Editing and image processing on a PGraphic. Using filters for improving picture clarity etc.
 *
 * features Filters for: contour filtering, contrast increasing, blue, highpass, 
 *                       lowpass, threshold, sharpen, relief, invert, greyscale and rgb-channel
 *
 *   also features the "optimize"-filter, which is a combination of a contrast (contour) image and thresholder. 
 *   It only runs through all pixels once, thus enabeling better performance, i.e. when doing Hough Transform preparation. 
 * 
 *   Always needs image to be set.  When setting image, a duplicate that won't be edited will be created, 
 *   so the user can still acces the original image. 
 *   All filter calling will affect the first image (img)
 *
 */


class GData {

  public PImage img;      //image to be processed
  public PImage uneditedImg;     //copy of image, so the original values can be recalled after editing
  int w;                  //width of image
  int h;                  //height of image
  public int param1 = 60; //image processing parameters
  public int param2 = 1;

OpenCV cv;

  GData() {
    
  }
  
  
  GData(PImage img) {
    this.img = img;
    this.w = img.width;
    this.h = img.height;
    this.uneditedImg = img;
  }

  GData(PApplet pa, int w, int h) {//parent, width, height
    cv = new OpenCV(pa, w, h);
  }
  void setImg(PImage imga) {   
    try {
      this.img = (PImage)imga.clone();
      this.uneditedImg = (PImage)imga.clone();
    }
    catch(Exception e) {
      e.printStackTrace();
    }    
    this.w = this.img.width;
    this.h = this.img.height;
  }


  void optimize() {
    colorMode(HSB, 255);
    int d = 2; //set distance to 2
    int t = 81; //set threshold to 81
    t = param1+1;
    //int t0 = millis();
    for (int i = width; i < this.img.pixels.length - d; i++) {
      this.img.pixels[i] = color(0, 0, min(255, 255*((int)(5*abs(brightness(this.img.pixels[i])-brightness(this.img.pixels[i+d])))/t)));
    }
    //println("Optimize by " + d + " time: " + (millis()-t0));
  }



  void contour(int d) { //d = strength
    colorMode(HSB, 255);
    for (int i = 0; i < this.img.pixels.length - d; i++) {
      this.img.pixels[i] = color(0, 0, 5*abs(brightness(this.img.pixels[i])-brightness(this.img.pixels[i+d])));
    }
    //println("Contour by " + d);
  }
  

  void advancedContour(int x0, int y0, int xMax, int yMax) {
    colorMode(RGB, 255);
    int time = millis();

    float[][] sobel = { //sobel matrix (better?)
      { 1, 0, -1 }, 
      { 2, 0, -2 }, 
      { 1, 0, -1 } 
    };
    float[][] laplace = { //lapacian operator
      {  0, -1, 0 }, 
      { -1, 4, -1 }, 
      {  0, -1, 0 } 
    };
    float[][] verticalEdges = { //only vertical edges
      {  0, -1, 0 }, 
      { -1, 4, -1 }, 
      {  0, -1, 0 } 
    };
    float[][] megaEdge = { //mega edge
      { -1, -1, -1 }, 
      { -1,  8, -1 }, 
      { -1, -1, -1 } 
    };
    float[][] kirsch = { //kirsch operator (g1)
      {  5,  5,  5 }, 
      { -3,  0, -3 }, 
      { -3, -3, -3 } 
    };
    
    //cv.loadImage(img);
    //cv.findCannyEdges(param1,param2);
    //cv.findScharrEdges(0);
    //cv.findSobelEdges(2, 5);
    //img = cv.getOutput();
    int mSize = 3;
    float factor =param1/200.0*10;
    x0 = constrain(x0, 0, this.img.width);
    y0 = constrain(y0, 0, this.img.height);
    xMax = constrain(xMax, 0, this.img.width);
    yMax = constrain(yMax, 0, this.img.height);
    for (int y = y0; y < yMax; y++) {
      for (int x = x0; x < xMax; x++) {
        this.img.pixels[x + y*this.img.width] = matrixConvolution(x, y, sobel, mSize, factor);
      }
    }
    //println("Contour t= " + (millis()-time));
  }

  color matrixConvolutionGrey(int x, int y, float[][] matrix, int mSize, float factor) {
    float newColor = 0;
    int offset = mSize/2;
    for (int j = 0; j < mSize; j++) {
      for (int i = 0; i < mSize; i++) {
        int location = constrain((x + i - offset) + this.img.width*(y+j-offset), 0, uneditedImg.pixels.length-1);
        newColor += (red(this.uneditedImg.pixels[location])+green(this.uneditedImg.pixels[location])+blue(this.uneditedImg.pixels[location]))/3 * matrix[j][i];
      }
    }
    newColor = constrain(abs(newColor)/factor, 0, 255);
    return color(newColor);
  }
  
  color matrixConvolution(int x, int y, float[][] matrix, int mSize, float factor) {
    float newRed = 0;
    float newGreen = 0;
    float newBlue = 0;
    int offset = mSize/2;
    for (int j = 0; j < mSize; j++) {
      for (int i = 0; i < mSize; i++) {
        int location = constrain((x + i - offset) + this.img.width*(y+j-offset), 0, uneditedImg.pixels.length-1);
        newRed += (red(this.uneditedImg.pixels[location]) * matrix[j][i]);
        newGreen += (green(this.uneditedImg.pixels[location]) * matrix[j][i]);
        newBlue += (blue(this.uneditedImg.pixels[location]) * matrix[j][i]);
      }
    }
    newRed = constrain(newRed/factor, 0, 255);
    newGreen = constrain(newGreen/factor, 0, 255);
    newBlue = constrain(newBlue/factor, 0, 255);
    return color(newRed, newGreen, newBlue);
  }


  void contrast(float strength, int threshold) { //increase contrast (works with inverted compressing)
    //strength = 12;
    //threshold = 120;
    colorMode(HSB, 255);
    for (int y = 0; y < this.img.pixels.length; y++) {
      float value = pow(brightness(this.img.pixels[y])/threshold, strength/10.0) * threshold;
      this.img.pixels[y] = color(0, 0, constrain(value, 0, 255));
    }
    //println("Contrast at " + threshold + "with slope of" + strength);
  }

  void blur() { //blur filter
    colorMode(RGB, 255);
    int time = millis();

    float[][] matrix = {
      { 4, 2, 4 }, 
      { 2, 1, 2 }, 
      { 4, 2, 4 } 
    };
    int mSize = 3;
    int factor = 25;

    for (int y = 0; y < this.img.height; y++) {
      for (int x = 0; x < this.img.width; x++) {
        this.img.pixels[x + y*this.img.width] = matrixConvolution(x, y, matrix, mSize, factor);
      }
    }
    //println("Blur " + (millis()-time) +" ms");
  }


  void sharpen() { //sharpening filter
    colorMode(RGB, 255);
    int time = millis();

    float[][] matrix = { 
      { -1, -1, -1 }, 
      { -1, 9, -1 }, 
      { -1, -1, -1 } 
    };
    int mSize = 3;
    float factor = .1;

    for (int y = 0; y < this.img.height; y++) {
      for (int x = 0; x < this.img.width; x++) {
        this.img.pixels[x + y*this.img.width] = matrixConvolution(x, y, matrix, mSize, factor);
      }
    }
    println("Sharpen " + (millis()-time));
  }









  void simpleBlur() { //very basic blur, only calculates 4 next neighbor pixels. can be used for very fast perfomance
    colorMode(HSB, 255);
    int time = millis();
    for (int i = this.w; i < this.img.pixels.length - 10 * this.w; i++) {
      this.img.pixels[i] = color(0, 0, (4*brightness(this.img.pixels[i])+2*(brightness(this.img.pixels[i+1])+brightness(this.img.pixels[i-1])+brightness(this.img.pixels[i-this.w])+brightness(this.img.pixels[i+this.w])))/8);
    }
    println("Blur " + (millis()-time) +" ms");
    this.img.updatePixels();
  }

  void lowPass(int threshold) {
    colorMode(HSB, 255);
    for (int y = 0; y < this.img.pixels.length; y++) {
      //this.img.pixels[y] = color(0, 0, min(threshold, (int)brightness(this.img.pixels[y])));
      this.img.pixels[y] = color(0, 0, min(threshold, (red(this.img.pixels[y])+blue(this.img.pixels[y])+green(this.img.pixels[y]))/3));
      //this.img.pixels[y] = color(min(threshold, red(this.img.pixels[y])), min(threshold, green(this.img.pixels[y])), min(threshold, blue(this.img.pixels[y])));
    }
    //println("LowPass " + threshold); 
    this.img.updatePixels();
  }

  void highPass(int threshold) {
    colorMode(HSB, 255);

    threshold = param1;
    for (int y = 0; y < this.img.pixels.length; y++) {
      this.img.pixels[y] = color(0, 0, max(threshold, (int)brightness(this.img.pixels[y])));
    }
    println("Highpass " + threshold); 
    this.img.updatePixels();
  }

  void notchFilter(int threshold) { 
    colorMode(HSB, 255);
    threshold = param1;
    for (int y = 0; y < this.h * this.w; y++) {
      float b = brightness(this.img.pixels[y]);
      if (b < threshold) {
        this.img.pixels[y] = color(0, 0, b);
      } else if (b > 255-threshold) {
        this.img.pixels[y] = color(0, 0, b);
      } else {
        this.img.pixels[y] = color(0, 0, 127);
      }
    }
  }

  void relief(int reliefHeight, int reliefBrightness) { //creates relief image; high contrast parts seem to be raised in height
    colorMode(HSB, 255);
    for (int i = 0; i < this.img.pixels.length - reliefHeight; i++) {
      this.img.pixels[i] = color(0, 0, reliefBrightness+brightness(this.img.pixels[i+reliefHeight])-brightness(this.img.pixels[i]));
    }
    println("Relief: Brightness " + reliefBrightness + ", Height: " + reliefHeight);
  }

  void sharpen(int d) { 
    colorMode(HSB, 255);
    for (int i = 0; i < this.img.pixels.length - d; i++) {
      this.img.pixels[i] = color(0, 0, 2*brightness(this.img.pixels[i+d])-brightness(this.img.pixels[i])); //keine klammern!
    }
    println("Sharpened by " + d);
  }

  void invert(int d) {
    colorMode(HSB, 255);
    for (int i = 0; i < this.img.pixels.length - d; i++) {
      this.img.pixels[i] = color(0, 0, 256-brightness(this.img.pixels[i]));
    }
    println("Inverted");
  }

  void threshold(int threshold) {
    colorMode(HSB, 255);
    this.img.loadPixels();
    for (int y = 0; y < this.img.pixels.length; y++) {
      this.img.pixels[y] = color(0, 0, min(255, 255*((int)brightness(this.img.pixels[y])/threshold)));
    }
    println("Thresholded at " + threshold);
    this.img.updatePixels();
  }

  void chromeFilter(int d) { //reminds of the photoshop filter "chrome"
    colorMode(HSB, 255);
    for (int i = 0; i < this.img.pixels.length - d; i++) {
      this.img.pixels[i] = color(0, 0, 100+30 * sin(1000/brightness(this.img.pixels[i])));
    }
    println("chrome");
  }

  void channel(int mode) { //mode 0: grey-scale image, 1: red channel, 2: green channel, 3: blue channel
    if (mode == 0) { //grey image (combines red, gree, blue)
      colorMode(HSB, 255);
      for (int i = 0; i < this.img.pixels.length; i++) {
        this.img.pixels[i] = color(0, 0, brightness(this.img.pixels[i]));
      }
      println("Gray-Scale");
    } else if (mode == 1) { //red channel
      colorMode(RGB, 255);
      for (int i = 0; i < this.img.pixels.length; i++) {
        int clr = (int)red(this.img.pixels[i]);
        this.img.pixels[i] = color(clr, clr, clr);
      }
      println("Red Channel");
    } else if (mode == 2) { //green channel
      colorMode(RGB, 255);
      for (int i = 0; i < this.img.pixels.length; i++) {
        int clr = (int)green(this.img.pixels[i]);
        this.img.pixels[i] = color(clr, clr, clr);
      }
      println("Green Channel");
    } else if (mode == 3) { //blue channel
      colorMode(RGB, 255);
      for (int i = 0; i < this.img.pixels.length; i++) {
        int clr = (int)blue(this.img.pixels[i]);
        img.pixels[i] = color(clr, clr, clr);
      }
      println("Blue Channel");
    }
  }
}
